from __future__ import annotations

from datetime import timedelta
from collections import defaultdict
import json
from logging import getLogger
from statistics import mean
from typing import Dict, List, Tuple, Optional, Iterable, Iterator, Union, TYPE_CHECKING
from pathlib import Path
import re
from random import Random

from IPython.display import display, Markdown
from markdown import markdown as to_html
from markdown_strings import esc_format as escape, link, header, bold, italics
from humanize import naturaldelta

from ..document import Document
from ..entity import Entity
from ..config import CONFIG as _CONFIG

if TYPE_CHECKING:  # avoid circular dependency
    from ..index import Results, Result, Index


LOGGER = getLogger(__name__)


Duration = timedelta
IndexName = str

SerializedResults = List[List]
EntityRepr = str

AllResults = Tuple[Dict[Entity, 'Results'], Duration, IndexName]
AllSerializedResults = List


def serialize(all_results: AllResults) -> AllSerializedResults:
    results, duration, index_name = all_results
    serialized_results = dict()
    for entity, result in results.items():
        entity_repr = repr(entity)
        serialized_results[entity_repr] = []
        for (document, (start, end)), similarity in result:
            test_snippet = document[start:end]
            serialized_result = [document.filename, start, end, test_snippet, similarity]
            serialized_results[entity_repr].append(serialized_result)
    serialized_duration = duration.total_seconds()
    all_serialized_results = [serialized_results, serialized_duration, index_name]
    return all_serialized_results


def deserialize(all_serialized_results: AllSerializedResults,
                entities: Optional[Iterable[Entity]]) -> AllResults:
    results = defaultdict(lambda: list())
    serialized_results, serialized_duration, index_name = all_serialized_results
    for entity_repr, serialized_result in serialized_results.items():
        entity = Entity.from_repr(entity_repr)
        for filename, start, end, test_snippet, similarity in serialized_result:
            document = Document(filename)
            assert document[start:end] == test_snippet
            result = ((document, (start, end)), similarity)
            results[entity].append(result)
        from ..index import sort_results
        results[entity] = sort_results(results[entity])
    if entities is not None:
        results = {entity: results[entity] for entity in entities}
    duration = timedelta(seconds=serialized_duration)
    all_results = (results, duration, index_name)
    return all_results


class SearchResultList:
    CONFIG = _CONFIG['search.SearchResultList']
    TOPN = CONFIG.getint('number_of_results')
    ROOT_PATH = Path(CONFIG['root_path'])

    def __init__(self, results: Dict[Entity, 'Results'], duration: Duration,
                 index: Union['Index', str]):
        self.results = dict(results)
        self.duration = duration
        self.index_name = str(index)

    def __getitem__(self, entity_or_entities: Union[Iterable[Entity], Entity]) -> Union['Results', 'SearchResultList']:
        if isinstance(entity_or_entities, Entity):
            entity = entity_or_entities
            result = self.results[entity]
            return result
        else:
            entities = entity_or_entities
            results = {entity: self[entity] for entity in entities}
            duration = self.duration * len(entities) / len(self)
            sliced_result_list = SearchResultList(results, duration, self.index_name)
            return sliced_result_list

    def __len__(self) -> int:
        return len(self.results)

    def __iter__(self) -> Iterator[Entity]:
        return iter(self.results)

    def filter_and_map_entities(self, entities: Dict[Entity, Entity]) -> 'SearchResultList':
        updated_results = dict()
        for entity in self:
            if entity not in entities:
                continue
            mapped_entity = entities[entity]
            updated_results[mapped_entity] = self[entity]
        return SearchResultList(updated_results, self.duration, self.index_name)

    def save_sentences_and_ner_tags(self, basename: str, *args, random_seed: int = 42, **kwargs) -> None:
        sentences_filename = self.__class__._get_sentences_filename(basename)
        ner_tags_filename = self.__class__._get_ner_tags_filename(basename)
        sentences_and_ner_tags = list(self.get_sentences_and_ner_tags(*args, **kwargs))
        Random(random_seed).shuffle(sentences_and_ner_tags)
        with sentences_filename.open('wt', encoding='utf8') as sf, \
                ner_tags_filename.open('wt', encoding='utf8') as ntf:
            for sentence, ner_tags in sentences_and_ner_tags:
                sentence = re.sub(r'\n+', ' ', sentence)
                assert '\n' not in sentence
                assert '\n' not in ner_tags
                print(sentence, file=sf)
                print(ner_tags, file=ntf)
        LOGGER.debug(f'Saved {sentences_filename} and {ner_tags_filename}')

    def get_sentences_and_ner_tags(self, only_relevant: bool, *args, **kwargs) -> Iterable[Tuple[str, str]]:
        for entity in self:
            for result in self[entity]:
                if TYPE_CHECKING:
                    assert isinstance(result, Result)
                result_head, _ = result
                document, position = result_head
                if only_relevant and not document.is_relevant:
                    continue

                snippet = document.get_sentence(position, *args, **kwargs)
                if snippet is None:
                    continue

                left_context, match, right_context = snippet
                left_context_tokens = left_context.split()
                match_tokens = match.split()
                right_context_tokens = right_context.split()

                ner_tags_list = (
                    len(left_context_tokens) * ['O'] +
                    len(match_tokens) * [entity.ner_tag] +
                    len(right_context_tokens) * ['O']
                )
                ner_tags = ' '.join(ner_tags_list)

                left_context_and_match_tokens = f'{left_context}{match}'.split()
                match_and_right_context_tokens = f'{match}{right_context}'.split()
                assert len(left_context_and_match_tokens) in (
                    len(left_context_tokens) + len(match_tokens),
                    len(left_context_tokens) + len(match_tokens) - 1,
                )
                assert len(match_and_right_context_tokens) in (
                    len(match_tokens) + len(right_context_tokens),
                    len(match_tokens) + len(right_context_tokens) - 1,
                )
                if len(left_context_and_match_tokens) == len(left_context_tokens) + len(match_tokens):
                    left_context_separator = ''
                else:
                    left_context_separator = ' '
                if len(match_and_right_context_tokens) == len(match_tokens) + len(right_context_tokens):
                    right_context_separator = ''
                else:
                    right_context_separator = ' '

                sentence = f'{left_context}{left_context_separator}{match}{right_context_separator}{right_context}'

                assert len(ner_tags.split()) == len(sentence.split())

                yield (sentence, ner_tags)

    @property
    def all_results(self) -> AllResults:
        return (self.results, self.duration, self.index_name)

    @classmethod
    def _get_json_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.json')
        return filename

    @classmethod
    def _get_html_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.html')
        return filename

    @classmethod
    def _get_sentences_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.sentences.txt')
        return filename

    @classmethod
    def _get_ner_tags_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.ner_tags.txt')
        return filename

    def save(self, basename: str) -> None:
        self._save_json(basename)
        self._save_html(basename)

    def _save_json(self, basename: str) -> None:
        filename = self.__class__._get_json_filename(basename)
        with filename.open('wt') as f:
            json.dump(serialize(self.all_results), f, indent=4, sort_keys=True)
        LOGGER.debug('Saved {}'.format(filename))

    def _save_html(self, basename: str) -> None:
        filename = self.__class__._get_html_filename(basename)
        with filename.open('wt', encoding='utf8') as f:
            print(self.html, file=f)
        LOGGER.debug('Saved {}'.format(filename))

    @classmethod
    def load(cls, basename: str, entities: Optional[Iterable[Entity]] = None) -> 'SearchResultList':
        filename = cls._get_json_filename(basename)
        with filename.open('rt') as f:
            all_results = deserialize(json.load(f), entities)
        LOGGER.debug('Loaded {}'.format(filename))
        results = SearchResultList(*all_results)
        results._save_html(basename)
        return results

    def __str__(self) -> str:
        lengths = list(map(len, self.results.values()))
        summary = 'Retrieved {} results for {} entities ({:.2f} on average, {} at minimum) in {} using {}.'
        summary = summary.format(sum(lengths), len(self), mean(lengths), min(lengths), naturaldelta(self.duration), self.index_name)
        return summary

    def __repr__(self) -> str:
        return str(self)

    def _ipython_display_(self) -> None:
        display(self.markdown)

    def __radd__(self, other) -> 'SearchResultList':
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __add__(self, other) -> 'SearchResultList':
        if isinstance(other, SearchResultList):
            assert not (self.results.keys() & other.results.keys())
            results = {**self.results, **other.results}
            duration = self.duration + other.duration
            index_name = 'Combined'
            combined_result_list = SearchResultList(results, duration, index_name)
            return combined_result_list
        raise NotImplementedError

    @property
    def markdown(self) -> Markdown:
        blocks = [str(self)]
        for entity in self:
            result = self[entity][:self.TOPN]
            if TYPE_CHECKING:
                assert isinstance(result, Results)
            top_line = 'Top {} result{} for the entity {}:'
            top_line = top_line.format(len(result), 's' if len(result) > 1 else '', bold(escape(entity)))
            items = []
            for result_number, ((document, position), similarity) in enumerate(result):
                snippet = document.get_markdown_snippet(position).data
                document_name = 'Document {}'.format(escape(repr(document)))
                document_link = link(document_name, document.get_snippet_url(position))
                document_link = bold('{}:'.format(document_link))
                score = italics('(score: {:.2f})'.format(similarity))
                item = '{}. {} {} {}'.format(result_number+1, document_link, snippet, score)
                items.append(item)
            interblock_separator = header('<a href="#{0}" id="{0}">{1}</a>'.format(entity, self.index_name), 1)
            block = '{}\n\n{}\n\n{}'.format(interblock_separator, top_line, '\n'.join(items))
            blocks.append(block)
        document = '\n\n'.join(blocks)
        return Markdown(document)

    @property
    def html(self) -> str:
        title = 'Named entity search results for {}'.format(self.index_name)
        body = to_html(self.markdown.data)
        return '<!DOCTYPE html>\n<meta charset="utf-8">\n<title>{}</title>\n{}'.format(title, body)
